module delve.compress;


//
// Provides Readers and Writers that compress and uncompress data.
// Primarily intended for use with buffered streams, they can also be used
// directly - although they don't add much value to std.zlib
// Compress and UnCompress, and don't provide their own buffering.
//
// These Readers and Writers are not designed for use with Selector.
// The intended use is with non-blocking readers/writers like
// files or memory buffers.
//


public import delve.io;

private {
    import std.zlib;
    static import alg = std.algorithm;
}


//
// local class to encapsulate a data buffer
//
private class Buffer {
    const(void)[] _data;
    size_t        _begin;

    this() {}

    // give data to the buffer, which must be empty
    void give(const(void)[] data) {
        assert(_begin == _data.length);
        _data = data;
        _begin = 0;
    }

    // remove some data from the buffer, putting it into the
    // given ptr and returning the number placed there, which can be 0
    size_t remove(void* ptr, size_t count) {
        size_t qty = alg.min(count, _data.length - _begin);
        if (qty) {
            ptr[0..qty] = _data[_begin.._begin+qty];
            _begin += qty;
        }
        return qty;
    }
}


//
// Uncompressing IReadable that gets its data from an IReadable.
//
public class UnCompressingReader : public IReadable {
    private {
        IReadable  _reader;
        UnCompress _unCompress;
        Buffer     _buffer;
        ulong      _max;
        ulong      _count;
    }

    // construct with specified max number of compressed bytes of data to read
    // from the underlying IReadable.
    public this(IReadable reader, ulong max) {
        _reader     = reader;
        _unCompress = new UnCompress();
        _buffer     = new Buffer();
        _max        = max;
        _count      = 0;
    }

    // read 1..count bytes into buffer from the underlying reader,
    // returning the number read or throwing an exception.
    // Will block if necessary to return at least 1 byte.
    public override size_t read(void* buffer, size_t count) {
        // try to get some buffered uncompressed data
        size_t qty = _buffer.remove(buffer, count);
        if (!qty) {
            if (_count < _max) {
                // get some compressed data (up to max) and uncompress it
                void[] data = new void[alg.min(16 * 1024, _max - _count)];
                size_t num = _reader.read(data.ptr, data.length);
                _count += num;
                data = data[0..num];
                _buffer.give(_unCompress.uncompress(data));
            }
            else {
                // all out of compressed data - flush _unCompress to wring out
                // the last of the uncompressed data
                _buffer.give(_unCompress.flush);
            }
            qty = _buffer.remove(buffer, count);
            if (!qty) {
                throw new EndOfFile("no more data");
            }
        }
        return qty;
    }
}


//
// Compressing IReadable that gets its data from an IReadable.
//
public class CompressingReader : public IReadable {
    private {
        IReadable _reader;
        Compress  _compress;
        Buffer    _buffer;
        ulong     _max;
        ulong     _count;
    }

    // construct with specified max number of uncompressed bytes of data to read
    // from the underlying IReadable.
    public this(IReadable reader, ulong max) {
        _reader   = reader;
        _compress = new Compress();
        _buffer   = new Buffer();
        _max      = max;
        _count    = 0;
    }

    // read 1..count bytes into buffer from the underlying reader,
    // returning the number read or throwing an exception.
    // Will block if necessary to return at least 1 byte.
    public override size_t read(void* buffer, size_t count) {
        // try to get some buffered uncompressed data
        size_t qty = _buffer.remove(buffer, count);
        if (!qty) {
            if (_count < _max) {
                // get some uncompressed data (up to max) and compress it
                void[] data = new void[alg.min(16 * 1024, _max - _count)];
                size_t num = _reader.read(data.ptr, data.length);
                _count += num;
                data = data[0..num];
                _buffer.give(_compress.compress(data));
            }
            else {
                // all out of uncompressed data - flush _compress to wring out
                // the last of the compressed data
                _buffer.give(_compress.flush);
            }
            qty = _buffer.remove(buffer, count);
            if (!qty) {
                throw new EndOfFile("no more data");
            }
        }
        return qty;
    }
}



//
// UnCompressing IWritable that gives its data to another IWriteable
//
public class UnCompressingWriter : public IWriteable{
    private {
        IWriteable _writer;
        UnCompress _unCompress;
        ulong      _count;
    }

    public this(IWriteable writer) {
        _writer     = writer;
        _unCompress = new UnCompress();
        _count      = 0;
    }

    private void write_all(in void[] data) {
        scope const(void)[] buf = data;
        while (buf.length) {
            size_t qty = _writer.write(buf.ptr, buf.length);
            _count += qty;
            buf = buf[qty..$];
        }
    }

    // flush the last of the uncompressed data to the underlying writer,
    // returning the total number of uncompressed bytes written.
    public ulong flush() {
        write_all(_unCompress.flush);
        return _count;
    }

    // uncompress some data and pass it on to underlying writer
    public override size_t write(in void* buffer, size_t count) {
        write_all(_unCompress.uncompress(buffer[0..count]));
        return count;
    }
}



//
// Compressing IWriteable that gives its data to another IWriteable.
//
public class CompressingWriter : public IWriteable {
    private {
        IWriteable _writer;
        Compress   _compress;
        ulong      _count;
    }

    public this(IWriteable writer) {
        _writer   = writer;
        _compress = new Compress();
        _count    = 0;
    }

    private void write_all(in void[] data) {
        scope const(void)[] buf = data;
        while (buf.length) {
            size_t qty = _writer.write(buf.ptr, buf.length);
            _count += qty;
            buf = buf[qty..$];
        }
    }

    // flush the last of the compressed data to the underlying writer,
    // returning the total number of compressed bytes written.
    public ulong flush() {
        write_all(_compress.flush);
        return _count;
    }

    // compress some data and pass it on to underlying writer
    public override size_t write(in void* buffer, size_t count) {
        write_all(_compress.compress(buffer[0..count]));
        return count;
    }
}





