module delve.io;

import delve.backtrace;
import delve.logging;

import std.path;
import std.file;
import std.conv;

import core.stdc.errno;
import core.sys.posix.unistd;
import core.sys.posix.poll;

import std.c.stdlib;
import std.c.stdio;
import std.c.string;

static import linux = std.c.linux.linux;


//----------------------------------------------------------------------
// Basic I/O
//----------------------------------------------------------------------

// exceptions

public abstract class IOException : Exception {
    this(string text) { super(text); }
}

// thrown when read or write transfers no data
public class EndOfFile : IOException {
    this(string text) { super(text); }
}

public class WriteException : IOException {
    this(string text) { super(text); }
}

public class ReadException : IOException {
    this(string text) { super(text); }
}

public class SeekException : IOException {
    this(string text) { super(text); }
}

public class FileException : IOException {
    this(string text) { super(text); }
}


// interfaces

public interface IReadable {
    // read 1..count bytes into buffer from the device,
    // returning the number read or throwing an exception.
    // Block if necessary to return at least 1 byte.
    size_t read(void* buffer, size_t count);
}

public interface IWriteable {
    // write 1..count bytes from buffer to the device,
    // returning the number written or throwing an exception.
    // Block if necessary to return at least 1 byte.
    size_t write(in void* buffer, size_t count);
}

interface ISelectableReadable : IReadable {
    protected int read_descriptor();
}

interface ISelectableWriteable : IWriteable {
    protected int write_descriptor();
}

// a read device that can seek
public interface ISeekableReadable : public IReadable {
    // seek to the specified position from the start of the reader,
    void seek(ulong new_position);

    // return the current position in the reader
    ulong position() const;
}

// a write device that knows its position
public interface IPositionWriteable : public IWriteable {
    // return the current position in the writer
    ulong position() const;
}


// helpers

public string errorMsg() {
    char[1024] buffer;
    const char* s = strerror_r(getErrno, buffer.ptr, buffer.length);
    size_t len = strlen(s);
    return s[0..len].idup;
}


// read from a file descriptor, retrying on EINTR and throwing on errors or EOF
public size_t fdRead(int fd, void* buffer, size_t count) {
    assert(fd >= 0);
    for (;;) {
        ssize_t rc = linux.read(fd, buffer, count);
        if (rc == -1) {
            if (getErrno == EINTR) {
                // interrupted - go around again
                continue;
            }
            else {
                throw new ReadException("read error: " ~ errorMsg);
            }
        }
        else if (rc == 0) {
            throw new EndOfFile("end of file on read");
        }
        return rc;
    }
}

// write to a file descriptor, retrying on EINTR and throwing on errors or EOF
public size_t fdWrite(int fd, in void* buffer, size_t count) {
    assert(fd >= 0);
    for (;;) {
        ssize_t rc = linux.write(fd, buffer, count);
        if (rc == -1) {
            if (getErrno == EINTR) {
                // interrupted - go around again
                continue;
            }
            else {
                throw new WriteException("write error: " ~ errorMsg);
            }
        }
        else if (rc == 0) {
            throw new EndOfFile("end of file on write");
        }
        return rc;
    }
}


//--------------------------------------------------------------------
// File I/O
//--------------------------------------------------------------------

private enum Whence {
    FromStart   = SEEK_SET,
    FromCurrent = SEEK_CUR,
    FromEnd     = SEEK_END
}

// Return a nul-terminated version of str
private immutable (char)* stringz(string str) {
    return (str ~ '\0').ptr;
}


//
// Abstract base file class with code reused by the others
//
public abstract class BaseFile {
    private {
        string _path;
        int    _fd = -1;
    }

    private this(string path, int fd) {
        _path = path;
        _fd = fd;
    }

    ~this() {
        close();
    }

    // Close the file
    void close() {
        if (_fd >= 0) {
            linux.close(_fd);
            _fd = -1;
        }
    }

    // Return this file's path
    string path() const {
        return _path;
    }

    // Return this file's size
    ulong size() const {
        return _path.getSize;
    }

    // Return the current position in this file
    private ulong _position() const {
        return cast(ulong) linux.lseek(_fd, 0, Whence.FromCurrent);
    }

    // Seek to the specified position, returning the position relative to the start of the file
    private ulong _seek(ulong offset, Whence whence) {
        assert(offset < int.max);
        assert(_fd != -1);
        long rc = linux.lseek(_fd, cast(int)offset, whence);
        if (rc < 0) {
            throw new Exception("seek error");
        }
        return cast(ulong) rc;
    }
}


//
// AppendFile - appends to a file
//
public class AppendFile : public BaseFile, public IPositionWriteable {

    // Create a File, opening the specified filesystem file for append.
    // If we are writing to a non-existant file, it is created.
    public this(string path) {
        int fd = linux.open(stringz(path), linux.O_WRONLY | linux.O_CREAT | linux.O_APPEND, octal!640);
        if (fd < 0) {
            throw new FileException("unable to open " ~ path ~ " for append");
        }
        super(path, fd);
    }

    // Truncate the file to the specified length
    public void truncate(ulong position) {
        assert(position < int.max);
        assert(_fd != -1);
        int rc = ftruncate(_fd, cast(int)position);
        if (rc == -1) {
            throw new FileException("unable to truncate " ~ _path);
        }
        _seek(0, Whence.FromEnd);
    }

    // Write all of the given buffer to file, or throw
    void write(in void[] buffer) {
        size_t next = 0;
        while (next < buffer.length) {
            size_t qty = write(buffer.ptr + next, buffer.length - next);
            next += qty;
        }
    }

    //
    // IPositionWriteable implementation
    //

    override size_t write(in void* buffer, size_t count) {
        return fdWrite(_fd, buffer, count);
    }

    override ulong position() const {
        return _position();
    }
}


//
// ReadFile - random-access read from a file
//
public class ReadFile : public BaseFile, public ISeekableReadable {

    // Create a File, opening the specified filesystem file for read.
    public this(string path) {
        int fd = linux.open(stringz(path), linux.O_RDONLY, octal!640);
        if (fd < 0) {
            throw new FileException("unable to open " ~ path ~ " for read");
        }
        super(path, fd);
    }

    // Read from the file into the given buffer, returning the populated slice or throwing.
    void[] read(void[] buffer) {
        size_t qty = read(buffer.ptr, buffer.length);
        return buffer[0..qty];
    }


    //
    // ISeekableReadable implementation
    //

    // Read some data from this file, returning number of bytes read or throwing.
    override size_t read(void* buffer, size_t count) {
        return fdRead(_fd, buffer, count);
    }

    // Seek to position from start of file
    public override void seek(ulong position) {
        _seek(position, Whence.FromStart);
    }

    // Return the current file position
    public override ulong position() const {
       return _position();
    }
}


//-------------------------------------------------------------------------------
// Selector - enables one thread to perform I/O with multiple Selectable objects,
// by identifying which objects can perform a read or write without blocking.
//-------------------------------------------------------------------------------

class SelectError : IOException {
    this(string text) { super(text); }
}

class Selector {
    alias void delegate(ISelectableReadable  reader, bool error) ReadableHandler;
    alias void delegate(ISelectableWriteable writer, bool error) WritableHandler;
    alias void delegate()                                        TimeoutHandler;

    private {
        struct ReadEntry {
            ISelectableReadable readable;
            ReadableHandler     handler;
        }

        struct WriteEntry {
            ISelectableWriteable writeable;
            WritableHandler      handler;
        }

        int             _timeout; // milliseconds
        TimeoutHandler  _timeoutHandler;
        ReadEntry[int]  _readEntries;
        WriteEntry[int] _writeEntries;
        pollfd[]        _pollFds;
        int             _numEvents;
        int             _index;
    }

    // Create a Selector without a timeout
    this() {
        _timeout = -1;
    }

    // Create a Selector with a timeout in seconds
    this(double timeout, TimeoutHandler handler) {
        _timeout        = cast(int)(timeout * 1000 + 0.5);
        _timeoutHandler = handler;
    }

    // Add or replace a read Handler.
    void addReadable(ISelectableReadable readable, ReadableHandler handler) {
        assert(readable);
        assert(handler);
        ReadEntry new_entry = { readable, handler };

        //log_trace("adding read handler on fd=%s", readable.read_descriptor);
        auto existing = readable.read_descriptor in _readEntries;
        if (existing) {
            *existing = new_entry;
        }
        else {
            _readEntries[readable.read_descriptor] = new_entry;
            pollfd entry;
            entry.fd      = readable.read_descriptor;
            entry.events  = POLLIN;
            entry.revents = 0;
            _pollFds    ~= entry;
        }
        _numEvents = 0;
    }

    // Add or replace a write Handler
    void addWriteable(ISelectableWriteable writeable, WritableHandler handler) {
        assert(writeable);
        assert(handler);
        WriteEntry new_entry = { writeable, handler };

        //log_trace("adding write handler on fd=%s", writeable.write_descriptor);
        auto existing = writeable.write_descriptor in _writeEntries;
        if (existing) {
            *existing = new_entry;
        }
        else {
            _writeEntries[writeable.write_descriptor] = new_entry;
            pollfd entry;
            entry.fd      = writeable.write_descriptor;
            entry.events  = POLLOUT;
            entry.revents = 0;
            _pollFds    ~= entry;
        }
        _numEvents = 0;
    }

    // Remove a read Handler
    void removeReadable(ISelectableReadable readable) {
        assert(readable);
        if (readable.read_descriptor in _readEntries) {
            //log_trace("removing read handler on fd=%s", readable.read_descriptor);
            _readEntries.remove(readable.read_descriptor);
            uint to = 0;
            foreach (uint from, pollfd pfd; _pollFds) {
                if ( from > to) {
                    _pollFds[to] = _pollFds[from];
                }
                if (!(pfd.fd == readable.read_descriptor && pfd.events == POLLIN)) {
                    ++to;
                }
            }
            _pollFds.length = _pollFds.length - 1;
            _numEvents = 0;
        }
    }

    // Remove a write Handler
    void removeWriteable(ISelectableWriteable writeable) {
        assert(writeable);
        if (writeable.write_descriptor in _writeEntries) {
            //log_trace("removing write handler on fd=%s", writeable.write_descriptor);
            _writeEntries.remove(writeable.write_descriptor);
            uint to = 0;
            foreach (uint from, pollfd pfd; _pollFds) {
                if (from > to) {
                    _pollFds[to] = _pollFds[from];
                }
                if (!(pfd.fd == writeable.write_descriptor && pfd.events == POLLOUT)) {
                    ++to;
                }
            }
            _pollFds.length = _pollFds.length - 1;
            _numEvents = 0;
        }
    }

    // Process the next I/O event or timeout, returning true if the
    // Selector still has handlers registered. Throws SelectError on error.
    public bool process() {
        linux.timeval timeout_val;
        if (_numEvents <= 0) {
            _numEvents = poll(_pollFds.ptr, _pollFds.length, _timeout);
            _index     = 0;
        }
        if (_numEvents < 0) {
            if (getErrno != EINTR) {
                throw new SelectError("selector error: " ~ errorMsg);
            }
        }
        else if (_numEvents == 0) {
            assert(_timeoutHandler);
            _timeoutHandler();
        }
        else {
            // skip over indexes without an event
            while (!_pollFds[_index].revents) {
                ++_index;
            }

            // _index has an event - call its handler
            short revents = _pollFds[_index].revents;
            short events  = _pollFds[_index].events;
            bool  error   = (revents & (POLLERR | POLLNVAL)) != 0;
            int   fd      = _pollFds[_index].fd;
            //log_trace("events=%s revents=%s fd=%s error=%s", events, revents, fd, error);
            if (events & POLLIN) {
                auto entry = fd in _readEntries;
                entry.handler(entry.readable, error);
            }
            else if (events & POLLOUT) {
                auto entry = fd in _writeEntries;
                entry.handler(entry.writeable, error);
            }
            else {
                assert(0, "logic error in Selector");
            }

            // advance to next index and return
            ++_index;
            --_numEvents;
        }
        return _readEntries.length > 0 || _writeEntries.length > 0;
    }
}


//-----------------------------------------------------------------------------------
// A Pipe for transferring raw data.
//-----------------------------------------------------------------------------------

class PipeReadable : ISelectableReadable {
    private int _fd;

    this(int fd) {
        _fd = linux.dup(fd);
    }

    ~this() {
        linux.close(_fd);
    }

    override size_t read(void* buffer, size_t count) {
        return fdRead(_fd, buffer, count);
    }

    override protected int read_descriptor() {
        return _fd;
    }
}

class PipeWriteable : ISelectableWriteable {
    private int _fd;

    this(int fd) {
        _fd = fd;
    }

    ~this() {
        linux.close(_fd);
    }

    override size_t write(in void* buffer, size_t count) {
        return fdWrite(_fd, buffer, count);
    }

    override protected int write_descriptor() {
        return _fd;
    }
}

synchronized class Pipe {
    private {
        int readFd, writeFd;
    }

    this() {
        int[2] fds;
        int rc = linux.pipe(fds);
        assert(!rc);
        readFd = fds[0];
        writeFd = fds[1];
    }

    ~this() {
        if (readFd  != -1) linux.close(readFd);
        if (writeFd != -1) linux.close(writeFd);
    }

    // one-shot return of a new reader
    PipeReadable readable() {
        assert(readFd != -1, "Can only get PipeReadable once");
        auto r = new PipeReadable(readFd);
        readFd = -1;
        return r;
    }

    // one-shot return of a new writer
    PipeWriteable writeable() {
        assert(writeFd != -1, "Can only get PipeWriteable once");
        auto w = new PipeWriteable(writeFd);
        writeFd = -1;
        return w;
    }
}

