module delve.logging;

import std.datetime;
import std.string;
import std.stdio;
import std.array;
import std.format;

import core.stdc.errno;
import core.stdc.string;

extern(C) int isatty(int fd);



//
// A simple thread-safe logging framework. Its has no local imports,
// so everything else can use it.
//
// Log entries should occupy one line only. They are written to stderr by default.
//

// public thread-local string used for thread name
public string myName = "no-name";

enum Severity { Trace, Info, Warning, Error, Fatal }

void setThreshold(Severity severity) {
    logger.setThreshold(severity);
}


//
// log a line of text with assorted severities
//

public void log_fatal(string file = __FILE__, int line = __LINE__, A...)(string fmt, A a) {
    logger.log(Severity.Fatal, file, line, myName, fmt, a);
    throw new Error("fatal error");
}

public void log_error(string file = __FILE__, int line = __LINE__, A...)(string fmt, A a) {
    logger.log(Severity.Error, file, line, myName, fmt, a);
}

public void log_warn(string file = __FILE__, int line = __LINE__, A...)(string fmt, A a) {
    logger.log(Severity.Warning, file, line, myName, fmt, a);
}

public void log_info(string file = __FILE__, int line = __LINE__, A...)(string fmt, A a) {
    logger.log(Severity.Info, file, line, myName, fmt, a);
}

public void log_trace(string file = __FILE__, int line = __LINE__, A...)(string fmt, A a) {
    logger.log(Severity.Trace, file, line, myName, fmt, a);
}


// return a string representation of errno
public string errorString() {
    auto errno = core.stdc.errno.getErrno;
    if (errno != 0) {
        char* desc = core.stdc.string.strerror(errno);
        return format("errno=%s %s", errno, desc[0..core.stdc.string.strlen(desc)]);
    }
    else {
        return "no error";
    }
}


// return up to n right-most chars of str
private const(char)[] right(const(char)[] str, uint n) nothrow {
    size_t pos = str.length < n ? 0 : str.length - n;
    return str[pos..$];
}


private class Logger {
    static immutable string[] severityStr = [ "Trace ", "Info  ", "Warn  ", "Error ", "Fatal "];

    Severity          threshold;
    bool              coloured;
    char[]            buffer;
    Appender!(char[]) writer;

    this(Severity threshold_) nothrow {
        threshold = threshold_;
        coloured  = isatty(2) == 1;
        buffer.length = 256;
        writer = Appender!(char[])(buffer);
    }

    // set the logging threshold
    void setThreshold(Severity severity) nothrow {
        threshold = severity;
    }

    void log(A...)(Severity severity, string file, int line, string threadName, string fmt, A a) {
        if (severity >= threshold) {
            writer.clear;
            formattedWrite(writer, "%s%-29.29s|%s|%-20.20s|%-10s|",
                           colourString(severity),
                           Clock.currTime.toISOExtString,
                           severityStr[severity],
                           right(format("%s:%s", file, line), 20),
                           right(threadName, 10));
            formattedWrite(writer, fmt, a);
            formattedWrite(writer, "%s", coloured ? "\033[0m" : "");
            stderr.writeln(writer.data);
        }
    } 

    string colourString(Severity severity) {
        if (coloured) {
            switch (severity) {
            case Severity.Trace:    return "\033[0m\033[32m";
            case Severity.Info:     return "\033[0m\033[34m";
            case Severity.Warning:  return "\033[0m\033[35m";
            case Severity.Error:    return "\033[0m\033[31m";
            case Severity.Fatal:    return "\033[0m\033[31m";
            default: assert(0, "unsupported severity");
            }
        }
        return "";
    }
}

private Logger logger;

static this() {
    logger = new Logger(Severity.Info);
}
