module delve.net;

public import delve.io;
public import delve.stream;

import delve.logging;

import std.string;
import std.conv;

import core.stdc.stdint;
import core.stdc.errno;

import core.sys.posix.arpa.inet;
import core.sys.posix.netinet.tcp;
import core.sys.posix.netinet.in_;
import core.sys.posix.sys.socket;
import core.sys.posix.unistd;
import core.sys.posix.fcntl;



//
// Provides basic TCP and UDP networking support.
//
// Does not use std.socket because that doesn't handle concurrency,
// and I wanted something more focused on what I needed.
//

struct addrinfo {
    int       ai_flags;
    int       ai_family;
    int       ai_socktype;
    int       ai_protocol;
    size_t    ai_addrlen;
    sockaddr* ai_addr;
    char*     ai_canonname;
    addrinfo* ai_next;
};

// For reference:
//
// struct sockaddr {
//     sa_family_t sa_family;
//     byte[14]    sa_data;
// }
//
// struct sockaddr_in {
//     sa_family_t sin_family; // ushort
//     in_port_t   sin_port;   // ushort
//     in_addr     sin_addr;
// }
//
// struct in_addr {
//     in_addr_t s_addr;       // uint
// }


extern (C) int getaddrinfo(const char*     node,
                           const char*     service,
                           const addrinfo* hints,
                           addrinfo**      res); // out parameter


public class NetException : IOException {
    this(string text) { super(text); }
}


// provides information about a named host.
struct HostInfo {
    string[] names;
    uint[]   addresses; // host byte order

    // hostname can be in IPV4 or IPV6 format, or a name
    this(string hostname) {
        addrinfo* info;
        addrinfo hints;
        hints.ai_family = AF_INET;

        int rc = getaddrinfo(toStringz(hostname), null, &hints, &info);
        if (rc) {
            throw new NetException(format("Could not find host %s", hostname));
        }

        while (info) {
            sockaddr_in* addr = cast(sockaddr_in*) info.ai_addr;

            names     ~= to!string(info.ai_canonname);
            addresses ~= ntohl(addr.sin_addr.s_addr);

            info = info.ai_next;
        }
    }
}


// An IPV4 address
struct Address {
    struct Addr {
	sockaddr_in sin;
        this(sockaddr_in sin_) { sin = sin_; }
    }
    private immutable(Addr)* addr;

    // Address that wraps the given sockaddr_in
    this(sockaddr_in sin) {
        addr = cast(immutable(Addr)*) new Addr(sin);
    }

    // Address suitable for TCP accepting on all interfaces on local host
    this(ushort port) {
        sockaddr_in sin;
        sin.sin_family      = AF_INET;
        sin.sin_port        = htons(port);
        sin.sin_addr.s_addr = 0;
        this(sin);
    }

    // Addresses suitable for connecting to
    this(string hostname, ushort port) {
        auto info = HostInfo(hostname);
        this(info.addresses[0], port);
    }
    this(uint host, ushort port) {
        sockaddr_in sin;
        sin.sin_family      = AF_INET;
        sin.sin_port        = htons(port);
        sin.sin_addr.s_addr = htonl(host);
        this(sin);
    }

    // streaming
    this(InStream stream) {
        sockaddr_in sin;
        stream.read(&sin, 1, sin.sizeof);
        this(sin);
    }
    void write(OutStream stream) const {
        stream.write(&addr.sin, 1, addr.sin.sizeof);
    }

    // Return the IPv4 address family
    ushort family() const {
        return ntohs(addr.sin.sin_family);
    }

    // Return the host's IP address
    uint host() const {
        return ntohl(addr.sin.sin_addr.s_addr);
    }

    // Return the IPv4 port number.
    ushort port() const {
        return ntohs(addr.sin.sin_port);
    }

    // Return true if the address is empty
    bool empty() const {
        return addr == null;
    }

    string toString() /*const*/ { // FIXME
        if (addr) {
            return
                to!string(inet_ntoa(addr.sin.sin_addr)) ~ ":" ~ 
                to!string(port);
        }
        else {
            return "null-address";
        }
    }

    //
    // methods needed to use Address as a key in an associative array
    //

    hash_t toHash() const {
        return (family * 0x100) + host + (addr.sin.sin_port * 0x10000);
    }

    bool opEquals(ref const Address other) const {
        return addr.sin == other.addr.sin;
    }

    int opCmp(ref const Address other) const {
        if (family != other.family) {
            return family - other.family;
        }
        else if (host != other.host) {
            return host - other.host;
        }
        else {
            return port - other.port;
        }
    }
}



//
// A connected TCP Socket
//
class TCPSocket: ISelectableReadable, ISelectableWriteable {

    private int _fd;

    // create a socket
    private this(int fd) {
        _fd = fd;
    }

    ~this() {
        close;
    }

    // shutdown the socket
    void shutdown() {
        .shutdown(_fd, SHUT_RDWR);
    }

    // close the socket
    void close() {
        if (_fd != -1) {
            shutdown;
            .close(_fd);
            _fd = -1;
        }
    }

    // return the local address of this socket
    Address localAddress() {
        sockaddr_in sin;
        uint length = cast(uint) sin.sizeof;
        int rc = .getsockname(_fd, cast(sockaddr*) &sin, &length);
        if (rc) throw new NetException("Failed to get local socket address");
        assert(length == sin.sizeof && sin.sin_family == AF_INET);
        return Address(sin);
    }

    // return the remote address of this socket
    Address remoteAddress() {
        sockaddr_in sin;
        uint length = cast(uint) sin.sizeof;
        int rc = .getpeername(_fd, cast(sockaddr*) &sin, &length);
        if (rc) throw new NetException("Failed to get remote socket address");
        assert(length == sin.sizeof && sin.sin_family == AF_INET);
        return Address(sin);
    }

    // return the value of a socket option
    int getOption(int level, int option) {
        int value;
        socklen_t len = value.sizeof;
        if (.getsockopt(_fd, level, option, &value, &len) < 0) {
            log_warn("failed to get socket option");
            return -1;
        }
        else {
            return value;
        }
    }

    // set a socket option, returning true on success
    bool setOption(int level, int option, int value) {
        socklen_t len = value.sizeof;
        if (.setsockopt(_fd, level, option, &value, len) < 0) {
            log_warn("failed to set socket option");
            return false;
        }
        else {
            return true;
        }
    }

     // set TCP Keepalive settings on the underlying socket
     bool setKeepalive(int idle, int interval, int count) {
         enum {
             SO_KEEPALIVE  = 9,
             TCP_KEEPIDLE  = 4,
             TCP_KEEPINTVL = 5,
             TCP_KEEPCNT   = 6
         }
         bool result = false;
         if (setOption(SOL_SOCKET, SO_KEEPALIVE, 1)) {                  // enable keepalive
             if (setOption(IPPROTO_TCP, TCP_KEEPIDLE, idle)) {          // idle secs before probe
                 if (setOption(IPPROTO_TCP, TCP_KEEPINTVL, interval)) { // secs between probes
                     if (setOption(IPPROTO_TCP, TCP_KEEPCNT, count)) {  // probes before dead
                         result = true;
                     }
                 }
             }
         }
         return result;
     }

    protected override int read_descriptor() {
        return _fd;
    }

    protected override int write_descriptor() {
        return _fd;
    }

    override size_t read(void* buffer, size_t count) {
        return fdRead(_fd, buffer, count);
    }

    override size_t write(in void* buffer, size_t count) {
        return fdWrite(_fd, buffer, count);
    }
}


//
// Connector - used to make a single connected client socket.
//
public class Connector : ISelectableWriteable {

    alias void delegate(Address, TCPSocket) Handler;

    private {
        Address  _address;
        Selector _selector;
        Handler  _handler;
        int      _fd;
    }

    // create a Connector, initiating a connection attempt and returning immediately
    public this(Address address, Selector selector, Handler handler) {
        _address  = address;
        _selector = selector;
        _handler  = handler;
        _fd       = .socket(AF_INET, SOCK_STREAM, 0);

        if (_fd == -1) throw new NetException("Unable to open TCP socket: " ~ errorString);
        //log_trace("Initiating connection to %s", _address);

        // set it to non-blocking
        int fflags = .fcntl(_fd, F_GETFL);
        fflags |= O_NONBLOCK;
        if (.fcntl(_fd, F_SETFL, fflags) == -1) {
            log_fatal("Failed to set non-blocking: " ~ errorString);
        }

        // initiate a connection attempt and restore to non-blocking
        sockaddr_in sin = address.addr.sin;
        if (.connect(_fd, cast(sockaddr*)(&sin), sin.sizeof) != -1) {
            log_fatal("Connect should return -1");
        }
        if (errno == EINPROGRESS) {
            fflags = .fcntl(_fd, F_GETFL);
            fflags &= ~O_NONBLOCK;
            if (.fcntl(_fd, F_SETFL, fflags) == -1) {
                log_fatal("Failed to unset non-blocking: " ~ errorString);
            }
        }
        else {
            .close(_fd);
            throw new NetException("Connect failed with: " ~ errorString);
        }

        _selector.addWriteable(this, &complete);
    }

    // Cancel a pending connection attempt
    void close() {
        if (_fd != -1) {
            _selector.removeWriteable(this);
            .close(_fd);
            _fd = -1;
        }
    }

    // Complete the connection attempt once _fd is writeable
    private void complete(ISelectableWriteable writeable, bool error) {
        assert(_fd != -1);
        _selector.removeWriteable(this);
        if (error) {
            //log_trace("Connect to %s failed", _address);
            _handler(_address, null);
        }
        else {
            auto socket = new TCPSocket(_fd);
            //log_trace("Connect to %s succeeded", _address);
            _handler(_address, socket);
        }
    }

    override size_t write(in void* buffer, size_t count) { assert(0, "invalid write attempt"); }
    override protected int write_descriptor() { return _fd; }
}


//
// Accepts incoming TCP connection requests, creating TCP sockets to service them.
//
class Acceptor {

    alias void delegate(TCPSocket) Handler;

    private {
        Address   _address;
        Selector  _selector;
        Handler   _handler;
        int       _fd;
        TCPSocket _socket;
    }

    // Create a socket bound and listening on address, which must be local.
    // Ports below 1024 require root privileges and even then the port may be occupied.
    this(Address address, Selector selector, Handler handler) {
        _address  = address;
        _selector = selector;
        _handler  = handler;

        _fd = .socket(AF_INET, SOCK_STREAM, 0);
        if (_fd == -1) log_fatal("failed to create socket");
        _socket = new TCPSocket(_fd);
        scope(failure) _socket.close;

        _socket.setOption(SOL_SOCKET, SO_REUSEADDR, 1);

        sockaddr_in sin = address.addr.sin;
        if (.bind(_fd, cast(sockaddr*)(&sin), sin.sizeof) == -1) {
            throw new NetException("Failed to bind to address");
        }

        if (.listen(_fd, 5) == -1) {
            throw new NetException("listen failed");
        }

        _selector.addReadable(_socket, &accept);
    }

    public void close() {
        if (_socket !is null) {
            _selector.removeReadable(_socket);
            _socket.close;
            _socket = null;
        }
    }

    private void accept(ISelectableReadable readable, bool error) {
        //log_trace("accepting an incoming connection");
        assert(!error);
        int fd = .accept(_fd, null, null);
        if (fd == -1) {
            //log_trace("accept error");
            throw new NetException("Accept failed");
        }
        auto socket = new TCPSocket(fd);
        //log_trace("accept succeeded fd=%s remoteAddress=%s", fd, socket.remoteAddress);
        _handler(socket);
    }
}


//
// A UDP Socket.
//
// Connectionless, unreliable datagrams.
// Each send sends a packet to the specified UDP address.
// Each recv returns the contents of a complete packet.
//
// Use by streaming data to/from Out/In MemoryStreams, and write/read
// the MemoryStream buffer.
//
// TODO - add multicast and broadcast support
// TODO - add streaming and selector support.
//

/*
class UDPSocket {

    private std.socket.Socket _socket;

    // create a UDP socket bound to local address
    this(ushort port) {
        _socket = new std.socket.Socket(std.socket.AddressFamily.INET,
                                        std.socket.SocketType.DGRAM,
                                        std.socket.ProtocolType.UDP);
        int val[1];
        val[0] = 1;
        _socket.setOption(std.socket.SocketOptionLevel.SOCKET,
                          std.socket.SocketOption.REUSEADDR,
                          val);

        // bind to the specified local port
        auto address = new Address(port);
        _socket.bind(address);
    }

    // write data to the specified destination
    void send(in void[] data, Address destination) {
        _socket.sendTo(data, destination);
    }

    // read data from anyone into buffer, returning received slice and setting origin.
    // return empty array if end of file.
    void[] recv(ubyte[] buffer, out Address origin) {
        int rc = _socket.receiveFrom(buffer, origin);
        return buffer[0..rc];
    }
}
*/
