module delve.stream;

public import delve.io;

import core.thread;

import std.c.string;
import std.typetuple;
import std.traits;
import std.algorithm;
import std.string;


//
// Provides facilities for I/O of binary data, including endian conversion.
//
// Stream I/O on devices is often potentially blocking, so precautions have
// to be taken to avoid stopping your program in its tracks.
// I/O of objects is also a complicated business, involving recursive calls
// into the stream get/put methods, and is usually only feasible if you use a stack
// to remember your state.
//
// The approaches to dealing with blocking I/O supported here are:
//
// * Stream in/out of memory streams and do blocking I/O with packets that are
//   simple enough so you don't need a stack to keep track of state.
//
// * Use Threads to perform I/O, one per device/direction, expecting them to block.
//   This approach maximises responsiveness and parallelism, but the threads can
//   be tricky to shut down, given that they can be blocked on I/O.
//
// * Use Fibres to perform I/O, one per device/direction, together with a Selector
//   to ensure that you never make a potentially blocking I/O call. This maximises
//   simplicity and per-cpu performance, and is recommended unless parallelism
//   is important.
//
// I/O operations may throw specialisations of IOException.
// Once an exception has been thrown by an InStream or OutStream it is
// invalid and no more methods may be invoked on it.
//


//----------------------------------------------------------------
// endian conversions
//----------------------------------------------------------------

//
// All data "outside" a host (on the wire, on disk, etc) should be in a
// specific byte-order, and this enum is used to specify that byte order.
// We support two byte-orders - big and little endian, with big-endian
// the strong preference.
//
public enum ByteOrder {
    BigEndian,
    LittleEndian
}

private static ByteOrder HostByteOrder;

static this() {
    ushort val = 0x1234;
    auto ptr = cast(ubyte*) &val;
    if (ptr[0] == 0x12)
        HostByteOrder = ByteOrder.BigEndian;
    else
        HostByteOrder = ByteOrder.LittleEndian;
}

//
// Copy multiple items, swapping the endian-ness of the items if they are longer
// than one byte and the host and specified byte-order are different.
// Endian conversion should be done exactly once on output and exactly once on
// input for each chunk of data.
//
private void endianCopy(void* dest, in void* src, size_t length, size_t width, ByteOrder byteOrder) {
    if (width != 1 && byteOrder != HostByteOrder) {
        auto d = cast(ubyte*) dest;
        auto s = cast(ubyte*) src;
        int s_index = 0;
        int d_index = 0;

        for(int index = 0; index < length; ++index) {
            s_index += width;
            for(int j = 0; j < width; ++j) {
                d[d_index++] = s[--s_index];
            }
            s_index += width;
        }
    }
    else {
        core.stdc.string.memcpy(dest, src, length * width);
    }
}


// convenience template for detecting a fundumental type
private template isFundamental(T) {
    enum bool isFundamental =
        is(T == enum) ||
        staticIndexOf!(Unqual!(T),
                 char, wchar, dchar,
                 bool,
                 byte, ubyte,
                 short, ushort,
                 int, uint,
                 long, ulong,
                 float, double, real) >= 0;
}


//---------------------------------------------------------------------
// Stream base classes
//---------------------------------------------------------------------

//
// Input stream
//
abstract class InStream {

    // Read specified number of items from the stream to fill the given
    // data buffer, doing endian conversion if required.
    // Length is the number of items, and Width is size of each item.
    // May throw IOException.
    abstract void read(void* data, size_t length, size_t width);

    // read something from the stream and return it
    final T get(T)() {
        static if (is(T == class)) {
            // assume it is a class that can be constructed from a stream
            return new T(this);
        }
        else static if (is(T == struct)) {
            // assume it is a struct that can be populated from a stream
            return T(this);
        }
        else static if (isFundamental!(T)) {
            // a fundumental type - read it
            T x;
            read(&x, 1, x.sizeof);
            return x;
        }
        else static if (is(T Value == Value[])) {
            // a dynamic array
            Value[] x;
            x.length = get!(uint)();
            static if (isFundamental!(Value)) {
                // array of fundumentals - read in one go
                read(cast(void *)x.ptr, x.length, Value.sizeof);
            }
            else {
                // array of non-fundumental - read one at a time
                for (size_t i = 0; i < x.length; ++i) {
                    x[i] = get!Value();
                }
            }
            return x;
        }
        else static if (is(T Value == Value[N], size_t N)) {
            // a static array
            Value[N] x;
            static if (isFundamental!(Value)) {
                // array of fundumentals - read in one go
                read(cast(void *)x.ptr, x.length, Value.sizeof);
            }
            else {
                // array of non-fundumental - read one at a time
                for (size_t i = 0; i < x.length; ++i) {
                    x[i] = get!Value();
                }
            }
            return x;
        }
        else static if (is(T Value == Value[Index], Index)) {
            // an associative array - read one element at a time
            size_t length = get!(uint)();
            Value[Index] x;
            for (size_t i = 0; i < length; ++i) {
                auto index = get!Index();
                x[index] = get!Value();
            }
            return x;
        }
        else {
            static assert(false, "read of " ~ T.stringof ~ " not supported");
        }
    }
}


//
// Output stream
//
abstract class OutStream {
    alias put opCall;

    // Write given data to the stream, doing endian conversion if required.
    // Length is the number of items, and width is size of each item.
    // May throw IOException.
    abstract void write(in void* data, size_t length, size_t width);

    // write an item to the stream
    final OutStream put(T)(in T x) {
        static if (is(T == class) || is(T == struct)) {
            // assume it is a class or struct with a write method
            x.write(this);
        }
        else static if (isFundamental!(T)) {
            // a fundumental type
            write(&x, 1, x.sizeof);
        }
        else static if (is(T Value : Value[])) {
            // dynamic array
            assert(x.length < uint.max, "array has too many elements to write to stream");
            uint length = cast(uint) x.length;
            put(length);
            static if (isFundamental!(Value)) {
                // a fundumental type - write in one go
                write(x.ptr, x.length, Value.sizeof);
            }
            else {
                // non-fundumental - write one at a time
                foreach (value; x) {
                    put(value);
                }
            }
        }
        else static if (is(T Value == Value[N], size_t N)) {
            // a static array
            static if (isFundamental!(Value)) {
                // a fundumental type - write in one go
                write(x.ptr, N, Value.sizeof);
            }
            else {
                // non-fundumental - write one at a time
                foreach (value; x) {
                    put(value);
                }
            }
        }
        else static if (is(T Value == Value[Index], Index)) {
            // an associative array - write one element at a time
            assert(x.length < uint.max, "assoc array has too many elements to write to stream");
            uint length = cast(uint) x.length;
            put(length);
            foreach (index, value; x) {
                put(index);
                put(value);
            }
        }
        else {
            static assert(false, "write of " ~ T.stringof ~ " not supported");
        }
        return this;
    }
}


//------------------------------------------------------------------
// Streams that work with memory buffers, and so cannot block
//------------------------------------------------------------------

//
// InStream from a memory buffer
//
class InMemoryStream: InStream {
    private {
        const ubyte[] _buffer;
        size_t        _index;
        ByteOrder     _byteOrder;
    }

    this(const ubyte[] buffer, ByteOrder byteOrder = ByteOrder.BigEndian) {
        _buffer = buffer;
        _byteOrder = byteOrder;
    }

    override void read(void* data, size_t length, size_t width) {
        size_t bytes = length * width;
        if (_index + bytes > _buffer.length) {
            throw new EndOfFile("attempt to read past end of buffer");
        }
        else {
            endianCopy(data, &_buffer[_index], length, width, _byteOrder);
            _index += bytes;
        }
    }
}

//
// OutStream to a memory buffer
//
class OutMemoryStream: OutStream {
    private {
        ubyte[]   _buffer;
        ByteOrder _byteOrder;
        size_t    _index;
    }

    this(size_t initial_size = 1024, ByteOrder byte_order = ByteOrder.BigEndian) {
        _buffer.length = initial_size;
        _byteOrder = byte_order;
    }

    // return the populated slice of the internal buffer
    ubyte[] data() {
        return _buffer[0 .. _index];
    }

    override void write(in void* data, size_t length, size_t width) {
        size_t bytes = length * width;
        while (_index + bytes > _buffer.length) {
            _buffer.length = _buffer.length * 2;
        }
        endianCopy(&_buffer[_index], data, length, width, _byteOrder);
        _index += bytes;
    }
}



//------------------------------------------------------------------
// Buffered streams working with IReadable and IWriteable I/O,
// which is potentially blocking.
//
// These streams are intended for use by Threads that are happy to
// block for extended periods.
//------------------------------------------------------------------


//
// Buffered InStream from an IReadable
//
class InBufferedStream : InStream {
    private {
        IReadable _readable;
        ubyte[]   _buffer;
        ByteOrder _byteOrder;
        size_t    _next;      // next index of buffer to read from
        size_t    _qty;       // number of unread bytes in buffer
    }

    this(IReadable readable,
         size_t    bufferSize = 4096,
         ByteOrder byteOrder  = ByteOrder.BigEndian)
    {
        _readable      = readable;
        _buffer.length = bufferSize;
        _byteOrder     = byteOrder;
    }

    // read all the specified items from the stream to data
    override void read(void* data, size_t length, size_t width) {
        assert(length && width && width <= _buffer.length,
               format("invalid buffered read: length=%s width=%s buffer.length=%s",
                      length, width, _buffer.length));
        auto dest = cast(ubyte*) data;
        size_t capacity = _buffer.length / width;

        size_t remaining = length;
        while (remaining > 0) {
            while (_qty < width) {
                // get all the data we can from the device, possibly blocking
                fill();
            }

            // we have enough in the buffer for at least one item
            size_t available = _qty / width;
            size_t items = min(available, remaining);

            // transfer some items to the caller's data buffer
            size_t bytes = items * width;
            endianCopy(dest, &_buffer[_next], items, width, _byteOrder);
            dest += bytes;
            remaining -= items;
            _next += bytes;
            _qty -= bytes;
        }
    }

    // fill the buffer with as much as we can get from a single read
    protected void fill() {
        // move remaining contents of buffer to beginning of buffer
        if (_qty > 0) {
            if (_next >= _qty) {
                // non-overlapping memory regions
                core.stdc.string.memcpy(&_buffer[0], &_buffer[_next], _qty);
            }
            else {
                // overlapping memory regions, must use memmove
                memmove(&_buffer[0], &_buffer[_next], _qty);
            }
        }
        _next = 0;

        // try to fill the buffer
        auto dest = &_buffer[_qty];
        size_t count = _readable.read(dest, _buffer.length - _qty);
        assert(count > 0);
        _qty += count;
    }
}


//
// Buffered OutStream to an IWriter
//
class OutBufferedStream : OutStream {
    private {
        IWriteable _writeable;
        ubyte[]    _buffer;
        ByteOrder  _byteOrder;
        size_t     _qty;
    }

    this(IWriteable writeable,
         size_t     buffer_size = 4096,
         ByteOrder  byte_order  = ByteOrder.BigEndian)
    {
        _writeable     = writeable;
        _buffer.length = buffer_size;
        _byteOrder     = byte_order;
    }

    // Write items to the stream.
    override void write(in void* data, size_t length, size_t width) {
        assert(width && width && width <= _buffer.length,
               format("invalid buffered write: length=%s width=%s buffer.length=%s",
                      length, width, _buffer.length));
        auto src = cast(ubyte*) data;
        size_t remaining = length;
        while (remaining > 0) {
            if ((_buffer.length - _qty) < width) {
                // not enough room in buffer - flush it all
                flush();
            }
            size_t room = (_buffer.length - _qty) / width;
            size_t items = min(remaining, room);
            endianCopy(&_buffer[_qty], src, items, width, _byteOrder);
            src += items * width;
            remaining -= items;
            _qty += items * width;
        }
    }

    // Flush the entire contents of the internal buffer to the underlying IWriter
    void flush() {
        ubyte* src = _buffer.ptr;
        while (_qty > 0) {
            // write some data to the underlying device
            size_t count = _writeable.write(src, _qty);
            assert(count);
            _qty -= count;
            src += count;
        }
    }
}



//------------------------------------------------------------------
// Extensions of buffered streams that keep track of position
// and provide seeking. 
//
// These streams are intended for use with files such as
// AppendFile and ReadFile
//
// They can exploit the buffered data when
// seeking, and understand the current position in the buffer's effect
// on the effective position in the file.
//------------------------------------------------------------------


//
// Input stream intended for random-access reading from a ReadFile.
// Note that the InSeekStream takes over management of the ISeekableReader
// and all seek calls must go through the InSeekStream.
//
class InSeekStream : public InBufferedStream {

    private {
        ulong             _bufferPosition; 
        ISeekableReadable _readable;
    }

    this(ISeekableReadable readable,
         size_t            buffer_size = 4096,
         ByteOrder         byte_order = ByteOrder.BigEndian)
    {
        super(readable, buffer_size, byte_order);
        _readable = readable;
        _bufferPosition = _readable.position;
    }

    // return the current position
    ulong position() {
        return _bufferPosition + _next;
    }

    // seek to position
    void seek(ulong new_position) {
        long diff = cast(long)new_position - cast(long)(_bufferPosition + _next);
        if (diff >= -cast(long)_next && diff < cast(long)_qty) {
            // movement is within the current buffer - just adjust the buffer
            _next += diff;
            _qty -= diff;
        }
        else {
            // clear the buffer and seek
            _next = 0;
            _qty = 0;
            _bufferPosition = new_position;
            _readable.seek(new_position);
        }
    }

    protected override void fill() {
        _bufferPosition += _next;
        super.fill();
    }
}


//
// Output stream intended for use with an AppendFile
//
class OutPositionStream : public OutBufferedStream {

    ulong _bufferPosition; 

    this(IPositionWriteable writeable,
         size_t             buffer_size = 4096,
         ByteOrder          byte_order = ByteOrder.BigEndian)
    {
        super(writeable, buffer_size, byte_order);
        _bufferPosition = writeable.position;
    }

    // return the current position
    ulong position() {
        return _bufferPosition + _qty;
    }

    override void flush() {
        _bufferPosition += _qty;
        super.flush();
    }
}


//-------------------------------------------------------------------------
// Extensions of buffered streams that cooperate with the caller
// and a Fiber to ensure that they never do potentially blocking I/O.
//
// These streams are intended for use by code that use Selector and Fiber.
// See the example below.
//-------------------------------------------------------------------------


//
// Buffered InStream from an IReadable using a Fiber to release the CPU
// when it might block on a read.
//
class InFiberStream : InBufferedStream {
    private {
        Fiber _fiber;
        bool  _armed;
        bool  _empty;
    }

    public this(Fiber     fiber,
                IReadable readable,
                size_t    buffer_size = 4096,
                ByteOrder byte_order  = ByteOrder.BigEndian)
    {
        assert(fiber);
        super(readable, buffer_size, byte_order);
        _fiber = fiber;
        _empty = true;
    }

    // tell the stream that its reader has data
    void arm() {
        _armed = true;
    }

    // return true if the stream has data
    bool hasData() {
        return _armed || !_empty;
    }

    // fill the buffer with as much as can be read immediately
    protected override void fill() {
        if (_armed) {
            // the reader has data - read some into buffer
            super.fill();
            _armed = false;
            _empty = false;
        }
        else {
            // can't be sure reader has data, and the buffer is definitely too empty
            _empty = true;
            _fiber.yield;
        }
    }
}


//
// Buffered OutStream to an IWriter using a Fiber to release the CPU
// when it might block on a write
//
class OutFiberStream : OutBufferedStream {
    private {
        Fiber _fiber;
        bool  _armed;
        bool  _full;
    }

    this(Fiber      fiber,
         IWriteable writer,
         size_t     buffer_size = 4096,
         ByteOrder  byte_order  = ByteOrder.BigEndian)
    {
        assert(fiber);
        super(writer, buffer_size, byte_order);
        _fiber = fiber;
        _full  = false;
    }

    // tell the stream that its writer can accept data
    void arm() {
        _armed = true;
    }

    // return true if the stream can accept more data without blocking
    bool hasRoom() {
        return _armed || !_full;
    }

    // Flush the entire contents of the internal buffer to the underlying IWriter,
    // yielding whenever we can't be sure the next write would block.
    override void flush() {
        ubyte* src = _buffer.ptr;
        while (_qty > 0) {
            // write some data to the underlying device
            if (_armed) {
                _armed = false;
                size_t count = _writeable.write(src, _qty);
                assert(count);
                _qty -= count;
                src += count;
                _full = false;
            }
            else {
                _fiber.yield;
            }
        }
    }
}
