import delve.io;
import delve.concurrency;
import delve.logging;

import std.stdio;
import std.string;
import std.conv;


alias Protocol!("Requests",    Message!("job",    string, "name")).code jobCode;
alias Protocol!("Results",     Message!("result", string, "job", string, "desc")).code resultCode;
alias Protocol!("Completions", Message!("done",   string, "name")).code doneCode;

mixin(jobCode);
mixin(resultCode);
mixin(doneCode);


//--------------------------------------
// planner
//--------------------------------------

class ResultHandler : Results.IHandler {
    private bool[string] _outstanding;

    void expect(string job) {
        _outstanding[job] = true;
    }

    bool allDone() {
        return _outstanding.length == 0;
    }

    override void result(string job, string desc) {
        log_info("got result %s: %s", job, desc);
        _outstanding.remove(job);
    }
}

void doPlanning(Requests requests, Results results, Completions completions) {
    auto handler = new ResultHandler();

    foreach (i; 0..10) {
        string job = "job " ~ to!string(i);
        log_info("sending %s", job);
        requests.job(job);
        handler.expect(job);
    }
    while (!handler.allDone) {
        results.receive(handler);
    }
    log_info("all jobs done - terminating");
    requests.finalize;
    completions.done(myName);
}


//------------------------------------
// worker
//------------------------------------

class JobHandler : Requests.IHandler {
    private Results results;

    this(Results results_) {
        results = results_;
    }

    override void job(string job) {
        log_info("got job %s", job);
        results.result(job, job ~ " result");
    }
}

void doWork(Requests requests, Results results, Completions completions) {
    auto handler = new JobHandler(results);

    try {
        for (;;) {
            requests.receive(handler);
        }
    }
    catch (ChannelFinalized) {}
    log_info("%s: channel finalized - terminating", myName);
    completions.done(myName);
}


//------------------------------------
// main
//------------------------------------

class MainHandler : Completions.IHandler {
    private uint _remaining;

    this(uint remaining) {
        _remaining = remaining;
    }

    bool allDone() {
        return _remaining == 0;
    }

    override void done(string name) {
        --_remaining;
        log_info("%s: %s is done - %s remaining", myName, name, _remaining);
    }
}

void main() {
    myName = "main";

    writefln("\n%s\n", jobCode);

    auto requests    = new Requests();
    auto results     = new Results();
    auto completions = new Completions();

    uint qty;

    spawn(&doPlanning, "planner", requests, results, completions);
    ++qty;
    foreach (i; 0..2) {
        spawn(&doWork, format("worker%s", i), requests, results, completions);
        ++qty;
    }

    auto handler = new MainHandler(qty);
    while (!handler.allDone) {
        completions.receive(handler);
    }
    log_info("%s: all child threads terminated - terminating", myName);
}
