module bedrock.maxim.test.net_test;

import delve.net;
import delve.io;
import delve.stream;
import delve.concurrency;
import delve.logging;


// Note - it is bad form to do potentially blocking I/O
//        in a selector callback. We do it here for simplicity
//        and because with the amount of data being transferred, there
//        is no chance to block.

void doServer(Address address, shared Channel!bool channel) {

    Selector selector;
    Acceptor acceptor;

    auto selectable = channel.newSelectable;

    // called when a TCP connection is accepted
    void accepted(TCPSocket socket) {
        log_info("got a client connection from %s", socket.remoteAddress);

        // read a string and write it back
        auto in_stream = new InBufferedStream(socket);
        auto text = in_stream.get!(string);
        log_info("got text: %s", text);
        auto out_stream = new OutBufferedStream(socket);
        out_stream(text);
        out_stream.flush;
        socket.close;
        log_info("closed socket");
    }

    // called when there is input on channel
    void finalized(ISelectableReadable reader, bool error) {
        log_info("finalized");
        selector.removeReadable(selectable);
        acceptor.close;
    }

    log_info("wiring");
    selector = new Selector();
    acceptor = new Acceptor(address, selector, &accepted);
    selector.addReadable(selectable, &finalized);

    while (!selector.process) {}

    log_info("terminated");
}

void doClient(Address address, shared Channel!bool channel) {

    Selector            selector;
    Connector           connector;
    ISelectableReadable selectable;


    // called when a connection is established
    void connected(Address addr, TCPSocket socket) {
        if (socket is null) {
            log_info("connection attempt failed - retrying soon");
        }
        else {
            log_info("connection established to %s", addr);

            // send and receive some text
            string text = "hello sailor";
            auto out_stream = new OutBufferedStream(socket);
            out_stream(text);
            out_stream.flush;
            log_info("wrote text - waiting for reply");
            auto in_stream = new InBufferedStream(socket);
            auto text2 = in_stream.get!(string);
            log_info("got text: %s", text2);
            assert(text2 == text);

            // cause all threads to terminate
            log_info("finalizing channel");
            channel.finalize;
        }
        connector = null;
    }

    // called on select timeout
    void timeout() {
        if (!connector) {
            log_info("retrying connection attempt");
            connector = new Connector(address, selector, &connected);
        }
    }

    // called when there is input on channel
    void finalized(ISelectableReadable readable, bool error) {
        log_info("finalized");
        selector.removeReadable(selectable);
    }

    log_info("wiring up");
    selector  = new Selector(100, &timeout);
    connector = new Connector(address, selector, &connected);
    selectable = channel.newSelectable;
    selector.addReadable(selectable, &finalized);

    log_info("processing events");
    while (!selector.process) {}

    log_info("terminated");
}


int main(char[][] args) {
    myName = "main";
    setThreshold(Severity.Trace);
    log_info("starting");
    try {
        auto channel = new shared(Channel!bool)();
        auto address = Address("localhost", 12345);

        log_info("starting server");
        spawn(&doServer, "server", address, channel);

        log_info("starting client");
        spawn(&doClient, "client", address, channel);

        log_info("waiting for channel to finalize");
        channel.remove;
    }
    catch (ChannelFinalized) {
        log_info("finalized");
    }
    catch (Exception ex) {
        log_fatal("unexpected exception %s", ex);
    }
    log_info("terminated");
    return 0;
}
