module delve.test.talker_test;

import delve.talker;
import delve.logging;

//
// A test of delve.talker, intended primarily as an example of how to use Talker.
//

alias Talker!string MyTalker;


class ClientHandler : MyTalker.Results.IHandler {
    private {
        Selector          _selector;
        MyTalker.Results  _results;
        MyTalker.Commands _commands;
        Address           _address;
        ChannelSelectable _selectable;
    }

    this(Selector          selector,
         MyTalker.Results  results,
         MyTalker.Commands commands,
         Address           address)
    {
        _selector   = selector;
        _results    = results;
        _commands   = commands;
        _address    = address;
        _selectable = results.newSelectable;

        _selector.addReadable(_selectable, &handleResult);
        _commands.connect(address);
    }

    void handleResult(ISelectableReadable readable, bool error) {
        _results.receive(this);
    }

    //
    // MyTalker.Results.IHandler overrides
    //

    override void connected(Address address) {
        log_info("connected - sending messages");
        _commands.send(address, "one");
        _commands.send(address, "two");
        _commands.send(address, "three");
        _commands.send(address, "final");
    }

    override void failed(Address address) {
        log_info("connect attempt failed - trying again");
        _commands.connect(address);
    }

    override void accepted(Address address) {
        log_fatal("got unexpected accept");
    }

    override void disconnected(Address address) {
        log_info("client lost connection to server - all done");
        _commands.terminate(false);
    }

    override void received(Address address, string msg) {
        log_info("client received '%s'", msg);
        if (msg == "final") {
            log_info("got final message - terminating the client talker");
            _commands.terminate(false);
        }
    }

    override void sent(Address address) {
        log_info("client sent to %s", address);
    }
}


class ServerHandler : MyTalker.Results.IHandler {
    private {
        Selector          _selector;
        MyTalker.Results  _results;
        MyTalker.Commands _commands;
        Address           _address;
        ChannelSelectable _selectable;
        bool              _done;
    }

    this(Selector          selector,
         MyTalker.Results  results,
         MyTalker.Commands commands)
    {
        _selector   = selector;
        _results    = results;
        _commands   = commands;
        _selectable = results.newSelectable;

        _selector.addReadable(_selectable, &handleResult);
    }

    void handleResult(ISelectableReadable readable, bool error) {
        _results.receive(this);
    }

    bool done() const { return _done; }

    //
    // MyTalker.Results.IHandler overrides
    //

    override void connected(Address address) {
        log_fatal("got unexpected connected()");
    }

    override void failed(Address address) {
        log_fatal("got unexpected connection failure");
    }

    override void accepted(Address address) {
        log_info("accepted connection from %s", address);
    }

    override void disconnected(Address address) {
        log_info("server lost connection to client - terminating the server talker");
        _commands.terminate(false);
        _done = true;
    }

    override void received(Address address, string msg) {
        log_info("server received '%s'", msg);
        _commands.send(address, msg);
    }

    override void sent(Address address) {
        log_info("server sent to %s", address);
    }
}


int main() {
    myName = "main";
    setThreshold(Severity.Trace);

    auto address = Address("localhost", 54321);
    auto dummy   = Address();

    auto serverResults = new MyTalker.Results();
    auto clientResults = new MyTalker.Results();

    auto serverCommands = new MyTalker.Commands();
    auto clientCommands = new MyTalker.Commands();

    log_info("spawning client and server");
    MyTalker.spawnTalker("client", dummy,   clientCommands, clientResults);
    MyTalker.spawnTalker("server", address, serverCommands, serverResults);

    log_info("creating client and server handlers");
    auto selector      = new Selector;
    auto clientHandler = new ClientHandler(selector, clientResults, clientCommands, address);
    auto serverHandler = new ServerHandler(selector, serverResults, serverCommands);

    for (;;) {
        selector.process;
        if (serverHandler.done) break;
    }

    log_info("all done");

    return 0;
}
