module delve.talker;

public import delve.net;
public import delve.stream;
public import delve.concurrency;

import delve.logging;

import core.thread;


//
// Provides tcp client/server communications using Fibers and a Selector
// to get all the I/O done without resorting to more than one additional Thread.
//
// The advantages of Fibers are that they are much faster at context switching,
// and avoid all the complicated, dangerous and slow synchronisation required
// by Threads, resulting in faster operation and simpler code.
//
// The disadvantage is that Fibers don't exploit multiple CPUs.
//
// The real purpose of this module is to provide a useful advanced example of
// how the the I/O framework in delve can be used. It brings together
// virtually everything else - concurrency, streaming, selecting, etc.
// In quite a modest amount of easy to understand code, you have a useful
// TCP communications mechanism - not bad at all.
//

template Talker(T) {

    // Feedback from a Talker
    alias Protocol!("Results",
                    Message!("connected",    Address, "address"),
                    Message!("failed",       Address, "address"),
                    Message!("accepted",     Address, "address"),
                    Message!("disconnected", Address, "address"),
                    Message!("sent",         Address, "address"),
                    Message!("received",     Address, "address", T, "msg")).code resultsCode;
    mixin(resultsCode);

    // Commands to a Talker
    alias Protocol!("Commands",
                    Message!("connect",    Address, "address"),
                    Message!("send",       Address, "address", T, "msg"),
                    Message!("sendToAll",  T,       "msg"),
                    Message!("disconnect", Address, "address"),
                    Message!("terminate",  bool,    "dummy")).code commandsCode;
    mixin(commandsCode);


    // Observer of Connection events.
    // At the accepting end, address is the address of the connecting client.
    // At the connecting client end, address is that of the acceptor.
    interface IConnectionObserver {
        void connected(Address address);        // an outgoing connection attempt succeeded
        void failed(Address address);           // an outgoing connection attempt failed
        void disconnected(Address address);     // a connection has been lost
        void sent(Address address);             // a message has been sent to a connection
        void received(Address address, T msg);  // a message has been received
    }


    private {

        //
        // Connection - handles I/O via a single connected TCP socket.
        //
        class Connection {
            Selector            _selector;
            IConnectionObserver _observer;
            Address             _address;
            //
            Connector           _connector;
            //
            TCPSocket           _socket;
            shared Channel!T    _queue;
            ChannelSelectable   _selectable;
            Fiber               _readFiber;
            Fiber               _writeFiber;
            InFiberStream       _input;
            OutFiberStream      _output;

            // connection initiated remotely
            public this(Selector            selector,
                        IConnectionObserver observer,
                        TCPSocket           socket)
            {
                _selector = selector;
                _observer = observer;
                _address  = socket.remoteAddress;

                consumate(socket);
            }

            // initiate connection attempt, returning immediately.
            this(Selector            selector,
                 IConnectionObserver observer,
                 Address             address)
            {
                _selector  = selector;
                _observer  = observer;
                _address   = address;
                _connector = new Connector(_address, _selector, &connected);
            }

            // handle completion of the connection attempt
            private void connected(Address address, TCPSocket socket) {
                _connector = null;
                log_trace("completing outgoing connection to %s", address);
                if (socket) {
                    // connection attempt succeeded
                    consumate(socket);
                    _observer.connected(_address);
                }
                else {
                    // connection attempt failed
                    _observer.failed(address);
                }
            }

            private void consumate(TCPSocket socket) {
                _socket     = socket;
                _queue      = cast(shared) new Channel!T;
                _selectable = _queue.newSelectable;
                _readFiber  = new Fiber(&readRun);
                _writeFiber = new Fiber(&writeRun);
                _input      = new InFiberStream(_readFiber, _socket);
                _output     = new OutFiberStream(_writeFiber, _socket);

                _selector.addReadable(_selectable, &messageToSend);
                _selector.addReadable(_socket,     &socketReadable);
            }

            //
            // commands
            //

            // return the address of the remote end of the connection
            Address address() {
                return _address;
            }

            // queue a message for sending
            void send(T msg) {
                _queue.add(msg);
            }

            // break the connection and unplug from the selector
            void disconnect() {
                if (_socket) {
                    _socket.shutdown;
                    _selector.removeReadable(_selectable);
                    _selector.removeReadable(_socket);
                    _selector.removeWriteable(_socket);
                    _socket = null;
                }
                else if (_connector) {
                    _connector.close;
                }
            }

            //
            // Selector callbacks
            //

            void socketReadable(ISelectableReadable readable, bool error) {
                if (error) {
                    bail("read error on socket");
                }
                else {
                    _input.arm;
                    try {
                        _readFiber.call;
                    }
                    catch (IOException ex)  {
                        bail("error reading message from socket");
                    }
                }
            }

            void messageToSend(ISelectableReadable readable, bool error) {
                _writeFiber.call;
            }

            void socketWriteable(ISelectableWriteable writeable, bool error) {
                if (error) {
                    bail("write error on socket");
                }
                else {
                    _output.arm;
                    try {
                        _writeFiber.call;
                    }
                    catch (IOException ex) {
                        bail("error writing message to socket");
                    }
                }
            }

            //
            // Fiber run delegates
            //

            void readRun() {
                for (;;) {
                    // read a message and deliver it
                    T msg = _input.get!(T);
                    _observer.received(_address, msg);
                }
            }

            void writeRun() {
                for (;;) {
                    // get a message from the queue
                    _writeFiber.yield;
                    T msg = _queue.remove;
                    _selector.removeReadable(_selectable);
                    _selector.addWriteable(_socket, &socketWriteable);

                    // send it
                    if (!_output.hasRoom) {
                        _writeFiber.yield;
                    }
                    _output.put(msg);
                    _output.flush;
                    _observer.sent(_address);
                    _selector.addReadable(_selectable, &messageToSend);
                    _selector.removeWriteable(_socket);
                }
            }

            //
            // Helpers
            //

            void bail(string text) {
                log_info("disconnecting due to %s", text);
                disconnect;
                _observer.disconnected(_address);
            }
        }


        //
        // Controller
        //
        class Controller : IConnectionObserver, Commands.IHandler {
            private {
                Commands            _commands;
                Results             _results;
                //
                Selector            _selector;
                Acceptor            _acceptor;
                Connection[Address] _connections;
                bool                _terminated;
            }

            public this(Address address, Commands commands, Results results) {
                _commands = commands;
                _results  = results;

                _selector = new Selector();
                if (!address.empty) {
                    log_trace("listening on address %s", address);
                    _acceptor = new Acceptor(address, _selector, &handleAccept);
                }
                _selector.addReadable(_commands.newSelectable, &handleCommand);
            }

            void run() {
                try {
                    for (;;) {
                        _selector.process();
                    }
                }
                catch (ChannelFinalized ex) {}
                terminate(false);
                log_trace("terminated");
            }

            //
            // Delegates for Selector and Acceptor events
            //

            void handleAccept(TCPSocket socket) {
                Address address = socket.remoteAddress;
                log_trace("accepted connection from %s", address);
                _connections[address] = new Connection(_selector, this, socket);
                _results.accepted(address);
            }

            void handleCommand(ISelectableReadable readable, bool error) {
                _commands.receive(this);
            }

            //
            // Handle Connection callbacks
            //

            override void failed(Address address) {
                log_trace("failed to connect to %s", address);
                _connections.remove(address);
                _results.failed(address);
            }

            override void connected(Address address) {
                log_trace("connected to %s", address);
                _results.connected(address);
            }

            override void sent(Address address) {
                _results.sent(address);
            }

            override void disconnected(Address address) {
                log_trace("lost connection to %s", address);
                _connections.remove(address);
                _results.disconnected(address);
            }

            override void received(Address address, T msg) {
                log_trace("received a message from %s", address);
                _results.received(address, msg);
            }

            //
            // Handle commands
            //

            override void connect(Address address) {
                _connections[address] = new Connection(_selector, this, address);
            }

            override void send(Address address, T msg) {
                auto connection = address in _connections;
                if (connection) {
                    connection.send(msg);
                }
                else {
                    log_warn("no connection to %s - cannot send", address);
                    _results.disconnected(address);
                }
            }

            override void sendToAll(T msg) {
                foreach (connection; _connections) {
                    connection.send(msg);
                }
            }

            override void disconnect(Address address) {
                auto connection = address in _connections;
                if (connection) {
                    connection.disconnect;
                    _connections.remove(address);
                }
            }

            override void terminate(bool dummy) {
                if (!_terminated) {
                    _terminated = true;
                    foreach (address, connection; _connections) {
                        connection.disconnect;
                    }
                    if (_acceptor) _acceptor.close;
                    _commands.finalize;
                }
            }
        }
    }

    // Spawned talker thread's run function
    void doTalking(Address  address,
                   Commands commands,
                   Results  results) {
        auto controller = new Controller(address, commands, results);
        controller.run;
    }

    // Spawn a talker thread to manage I/O of a single data type T over TCP.
    void spawnTalker(string   name,
                     Address  address,
                     Commands commands,
                     Results  results) {
        spawn(&doTalking, name, address, commands, results);
    }
}
